#include <windows.h>
#include <stdio.h>

void print_help() {
    printf("Usage:\n"
           "  screcover [-<switches>] <source file> <target file>\n"
           "Switches:\n"
           "  -b <blocksize>  sets physical block size for media. 2048 is default\n"
          );
    exit(0);
}

__int64
get_file_size(
    HANDLE h
    )
{
    LONG hsz = 0;
    LONG lsz;

    lsz = SetFilePointer(h, 0, &hsz, FILE_END);
    return (((__int64)hsz) << 32) | lsz;
}

__int64
set_file_pointer(
    HANDLE h,
    __int64 sz
    )
{
    ULONG hsz = (ULONG)(sz >> 32);
    ULONG lsz = (ULONG)sz;

    lsz = SetFilePointer(h, lsz, (PLONG)&hsz, FILE_BEGIN);
    return (((__int64)hsz) << 32) | lsz;
}

int
set_file_size(
    HANDLE h,
    __int64 sz
    )
{
    set_file_pointer(h, sz);
    return SetEndOfFile(h);
}

void main (int argc, char* argv[]) {
    int i, j;
    char a;
    ULONG bs = 2048;
    char* in_name = NULL;
    char* out_name = NULL;
    int break_sw;
    __int64 f_size;
    __int64 f_ptr;
    __int64 bads = 0;
    char* buff;
    ULONG read_bytes;

    HANDLE in_h;
    HANDLE out_h;

    printf("Scratched File Copy utility for Windows, v.1\n"
           "Copyright (c) Alexander A. Telyatnikov, 2003\n");

    if(argc < 3)
        print_help();
    for(i=1; i<argc; i++) {
        if(!argv[i])
            continue;
        if(argv[i][0] != '-') {
            if(!in_name) {
                in_name = argv[i];
                continue;
            } else
            if(!out_name) {
                out_name = argv[i];
                continue;
            } else {
                print_help();
            }
        }
        j=1;
        break_sw = 0;
        while(!break_sw &&
              (a = argv[i][j]) && (a != ' ') && (a != '\t')) {
            switch(a) {
            case 'b' :
                i++;
                if(!argv[i])
                    print_help();
                sscanf(argv[i], "%d", &bs);
                break_sw = 1;
                break;
            case '?' :
            default:
                print_help();
            }
            j++;
        }
    }
    if(!in_name || !out_name)
        print_help();

    in_h = CreateFile(in_name, GENERIC_READ, FILE_SHARE_READ, NULL,
                      OPEN_EXISTING,
                      FILE_ATTRIBUTE_NORMAL /*| FILE_FLAG_NO_BUFFERING*/ | FILE_FLAG_SEQUENTIAL_SCAN,
                      NULL);
    if(in_h == (HANDLE)(-1)) {
        printf("Can't open source file %s\n"
               "exiting...\n", in_name);
        exit (-1);
    }

    buff = (char*)GlobalAlloc(GMEM_DISCARDABLE, bs);
    if(!buff) {
        printf("Can't alloc %d bytes for temporary buffer\n"
               "exiting...\n", bs);
        exit (-2);
    }

    out_h = CreateFile(out_name, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ,
                      NULL,
                      OPEN_ALWAYS,
                      FILE_ATTRIBUTE_NORMAL /*| FILE_FLAG_NO_BUFFERING*//* | FILE_FLAG_SEQUENTIAL_SCAN*/,
                      NULL);
    if(out_h == (HANDLE)(-1)) {
        read_bytes = GetLastError();
        printf("Can't create target file %s\n"
               "exiting...\n", out_name);
        exit (-3);
    }

    f_size = get_file_size(in_h);
    if(!f_size) {
        printf("Empty source file %s\n"
               "exiting...\n", out_name);
        exit (-4);
    }

    if(!set_file_size(out_h, f_size)) {
        printf("Can't set size of target file %s to %I64d bytes\n"
               "exiting...\n", out_name, f_size);
        exit (-5);
    }

    f_ptr = 0;

    while(f_ptr <= f_size-bs) {
        set_file_pointer(in_h, f_ptr);
        set_file_pointer(out_h, f_ptr);
        read_bytes = 0;
        memset(buff, 0, bs);
        ReadFile(in_h, buff, bs, &read_bytes, NULL);
        if(read_bytes != bs) {
            read_bytes = GetLastError();
            printf("Can't read block at offset %I64d\n", f_ptr);
            bads += bs;
        } else {
            WriteFile(out_h, buff, bs, &read_bytes, NULL);
            if(read_bytes != bs) {
                printf("Can't write block at offset %I64d\n", f_ptr);
            }
        }
        f_ptr += bs;
    }

    bs = (ULONG)(min(bs, f_size - f_ptr));
    if(!bs) {
        printf("Copied %I64d bytes, %I64d in bad blocks\n", f_size, bads);
        exit(0);
    }
    memset(buff, 0, bs);
    ReadFile(in_h, buff, bs, &read_bytes, NULL);
    if(read_bytes != bs) {
        printf("Can't read block at offset %I64d\n", f_ptr);
        bads += bs;
    }
    WriteFile(out_h, buff, bs, &read_bytes, NULL);
    if(read_bytes != bs) {
        printf("Can't write block at offset %I64d\n", f_ptr);
    }

    printf("Copied %I64d bytes, %I64d in bad blocks\n", f_size, bads);
    exit(0);
}
